# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Tests for the I{hosts(5)}-based resolver, L{twisted.names.hosts}.
"""
from __future__ import annotations

from typing_extensions import Protocol

from twisted.internet.defer import gatherResults
from twisted.names.dns import (
    AAAA,
    IN,
    A,
    DomainError,
    Query,
    Record_A,
    Record_AAAA,
    RRHeader,
)
from twisted.names.hosts import Resolver, searchFileFor, searchFileForAll
from twisted.python.filepath import FilePath
from twisted.trial.unittest import SynchronousTestCase


class _SupportsMktemp(Protocol):
    def mktemp(self) -> str:
        ...


class GoodTempPathMixin:
    def path(self: _SupportsMktemp) -> FilePath[bytes]:
        return FilePath(self.mktemp().encode("utf-8"))


class SearchHostsFileTests(SynchronousTestCase, GoodTempPathMixin):
    """
    Tests for L{searchFileFor}, a helper which finds the first address for a
    particular hostname in a I{hosts(5)}-style file.
    """

    def test_findAddress(self) -> None:
        """
        If there is an IPv4 address for the hostname passed to L{searchFileFor},
        it is returned.
        """
        hosts = self.path()
        hosts.setContent(b"10.2.3.4 foo.example.com\n")
        self.assertEqual("10.2.3.4", searchFileFor(hosts.path, b"foo.example.com"))

    def test_notFoundAddress(self) -> None:
        """
        If there is no address information for the hostname passed to
        L{searchFileFor}, L{None} is returned.
        """
        hosts = self.path()
        hosts.setContent(b"10.2.3.4 foo.example.com\n")
        self.assertIsNone(searchFileFor(hosts.path, b"bar.example.com"))

    def test_firstAddress(self) -> None:
        """
        The first address associated with the given hostname is returned.
        """
        hosts = self.path()
        hosts.setContent(
            b"::1 foo.example.com\n"
            b"10.1.2.3 foo.example.com\n"
            b"fe80::21b:fcff:feee:5a1d foo.example.com\n"
        )
        self.assertEqual("::1", searchFileFor(hosts.path, b"foo.example.com"))

    def test_searchFileForAliases(self) -> None:
        """
        For a host with a canonical name and one or more aliases,
        L{searchFileFor} can find an address given any of the names.
        """
        hosts = self.path()
        hosts.setContent(
            b"127.0.1.1\thelmut.example.org\thelmut\n"
            b"# a comment\n"
            b"::1 localhost ip6-localhost ip6-loopback\n"
        )
        self.assertEqual(searchFileFor(hosts.path, b"helmut"), "127.0.1.1")
        self.assertEqual(searchFileFor(hosts.path, b"helmut.example.org"), "127.0.1.1")
        self.assertEqual(searchFileFor(hosts.path, b"ip6-localhost"), "::1")
        self.assertEqual(searchFileFor(hosts.path, b"ip6-loopback"), "::1")
        self.assertEqual(searchFileFor(hosts.path, b"localhost"), "::1")


class SearchHostsFileForAllTests(SynchronousTestCase, GoodTempPathMixin):
    """
    Tests for L{searchFileForAll}, a helper which finds all addresses for a
    particular hostname in a I{hosts(5)}-style file.
    """

    def test_allAddresses(self) -> None:
        """
        L{searchFileForAll} returns a list of all addresses associated with the
        name passed to it.
        """
        hosts = self.path()
        hosts.setContent(
            b"127.0.0.1     foobar.example.com\n"
            b"127.0.0.2     foobar.example.com\n"
            b"::1           foobar.example.com\n"
        )
        self.assertEqual(
            ["127.0.0.1", "127.0.0.2", "::1"],
            searchFileForAll(hosts, b"foobar.example.com"),
        )

    def test_caseInsensitively(self) -> None:
        """
        L{searchFileForAll} searches for names case-insensitively.
        """
        hosts = self.path()
        hosts.setContent(b"127.0.0.1     foobar.EXAMPLE.com\n")
        self.assertEqual(["127.0.0.1"], searchFileForAll(hosts, b"FOOBAR.example.com"))

    def test_readError(self) -> None:
        """
        If there is an error reading the contents of the hosts file,
        L{searchFileForAll} returns an empty list.
        """
        self.assertEqual([], searchFileForAll(self.path(), b"example.com"))

    def test_malformedIP(self) -> None:
        """
        L{searchFileForAll} ignores any malformed IP addresses associated with
        the name passed to it.
        """
        hosts = self.path()
        hosts.setContent(
            b"127.0.0.1\tmiser.example.org\tmiser\n"
            b"not-an-ip\tmiser\n"
            b"\xffnot-ascii\t miser\n"
            b"# miser\n"
            b"miser\n"
            b"::1 miser"
        )
        self.assertEqual(
            ["127.0.0.1", "::1"],
            searchFileForAll(hosts, b"miser"),
        )


class HostsTests(SynchronousTestCase, GoodTempPathMixin):
    """
    Tests for the I{hosts(5)}-based L{twisted.names.hosts.Resolver}.
    """

    def setUp(self) -> None:
        f = self.path()
        f.setContent(
            b"""
1.1.1.1    EXAMPLE EXAMPLE.EXAMPLETHING
::2        mixed
1.1.1.2    MIXED
::1        ip6thingy
1.1.1.3    multiple
1.1.1.4    multiple
::3        ip6-multiple
::4        ip6-multiple
not-an-ip  malformed
malformed
# malformed
1.1.1.5    malformed
::5        malformed
"""
        )
        self.ttl = 4200
        self.resolver = Resolver(f.path, self.ttl)

    def test_defaultPath(self) -> None:
        """
        The default hosts file used by L{Resolver} is I{/etc/hosts} if no value
        is given for the C{file} initializer parameter.
        """
        resolver = Resolver()
        self.assertEqual(b"/etc/hosts", resolver.file)

    def test_getHostByName(self) -> None:
        """
        L{hosts.Resolver.getHostByName} returns a L{Deferred} which fires with a
        string giving the address of the queried name as found in the resolver's
        hosts file.
        """
        data = [
            (b"EXAMPLE", "1.1.1.1"),
            (b"EXAMPLE.EXAMPLETHING", "1.1.1.1"),
            (b"MIXED", "1.1.1.2"),
        ]
        ds = [
            self.resolver.getHostByName(n).addCallback(self.assertEqual, ip)
            for n, ip in data
        ]
        self.successResultOf(gatherResults(ds))

    def test_lookupAddress(self) -> None:
        """
        L{hosts.Resolver.lookupAddress} returns a L{Deferred} which fires with A
        records from the hosts file.
        """
        d = self.resolver.lookupAddress(b"multiple")
        answers, authority, additional = self.successResultOf(d)
        self.assertEqual(
            (
                RRHeader(b"multiple", A, IN, self.ttl, Record_A("1.1.1.3", self.ttl)),
                RRHeader(b"multiple", A, IN, self.ttl, Record_A("1.1.1.4", self.ttl)),
            ),
            answers,
        )

    def test_lookupIPV6Address(self) -> None:
        """
        L{hosts.Resolver.lookupIPV6Address} returns a L{Deferred} which fires
        with AAAA records from the hosts file.
        """
        d = self.resolver.lookupIPV6Address(b"ip6-multiple")
        answers, authority, additional = self.successResultOf(d)
        self.assertEqual(
            (
                RRHeader(
                    b"ip6-multiple", AAAA, IN, self.ttl, Record_AAAA("::3", self.ttl)
                ),
                RRHeader(
                    b"ip6-multiple", AAAA, IN, self.ttl, Record_AAAA("::4", self.ttl)
                ),
            ),
            answers,
        )

    def test_lookupAllRecords(self) -> None:
        """
        L{hosts.Resolver.lookupAllRecords} returns a L{Deferred} which fires
        with A records from the hosts file.
        """
        d = self.resolver.lookupAllRecords(b"mixed")
        answers, authority, additional = self.successResultOf(d)
        self.assertEqual(
            (RRHeader(b"mixed", A, IN, self.ttl, Record_A("1.1.1.2", self.ttl)),),
            answers,
        )

    def test_notImplemented(self) -> None:
        """
        L{hosts.Resolver} fails with L{NotImplementedError} for L{IResolver}
        methods it doesn't implement.
        """
        self.failureResultOf(
            self.resolver.lookupMailExchange(b"EXAMPLE"), NotImplementedError
        )

    def test_query(self) -> None:
        d = self.resolver.query(Query(b"EXAMPLE"))
        [answer], authority, additional = self.successResultOf(d)
        self.assertEqual(answer.payload.dottedQuad(), "1.1.1.1")

    def test_lookupAddressNotFound(self) -> None:
        """
        L{hosts.Resolver.lookupAddress} returns a L{Deferred} which fires with
        L{dns.DomainError} if the name passed in has no addresses in the hosts
        file.
        """
        self.failureResultOf(self.resolver.lookupAddress(b"foueoa"), DomainError)

    def test_lookupIPV6AddressNotFound(self) -> None:
        """
        Like L{test_lookupAddressNotFound}, but for
        L{hosts.Resolver.lookupIPV6Address}.
        """
        self.failureResultOf(self.resolver.lookupIPV6Address(b"foueoa"), DomainError)

    def test_lookupAllRecordsNotFound(self) -> None:
        """
        Like L{test_lookupAddressNotFound}, but for
        L{hosts.Resolver.lookupAllRecords}.
        """
        self.failureResultOf(self.resolver.lookupAllRecords(b"foueoa"), DomainError)

    def test_lookupMalformed(self) -> None:
        """
        L{hosts.Resolver.lookupAddress} returns a L{Deferred} which fires with
        the valid addresses from the hosts file, ignoring any entries that
        aren't valid IP addresses.
        """
        d = self.resolver.lookupAddress(b"malformed")
        [answer], authority, additional = self.successResultOf(d)
        self.assertEqual(
            RRHeader(b"malformed", A, IN, self.ttl, Record_A("1.1.1.5", self.ttl)),
            answer,
        )

    def test_lookupIPV6Malformed(self) -> None:
        """
        Like L{test_lookupAddressMalformed}, but for
        L{hosts.Resolver.lookupIPV6Address}.
        """
        d = self.resolver.lookupIPV6Address(b"malformed")
        [answer], authority, additional = self.successResultOf(d)
        self.assertEqual(
            RRHeader(b"malformed", AAAA, IN, self.ttl, Record_AAAA("::5", self.ttl)),
            answer,
        )



# -*- test-case-name: twisted.application.twist.test.test_twist -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

"""
Run a Twisted application.
"""

import sys
from typing import Sequence

from twisted.application.app import _exitWithSignal
from twisted.internet.interfaces import IReactorCore, _ISupportsExitSignalCapturing
from twisted.python.usage import Options, UsageError
from ..runner._exit import ExitStatus, exit
from ..runner._runner import Runner
from ..service import Application, IService, IServiceMaker
from ._options import TwistOptions


class Twist:
    """
    Run a Twisted application.
    """

    @staticmethod
    def options(argv: Sequence[str]) -> TwistOptions:
        """
        Parse command line options.

        @param argv: Command line arguments.
        @return: The parsed options.
        """
        options = TwistOptions()

        try:
            options.parseOptions(argv[1:])
        except UsageError as e:
            exit(ExitStatus.EX_USAGE, f"Error: {e}\n\n{options}")

        return options

    @staticmethod
    def service(plugin: IServiceMaker, options: Options) -> IService:
        """
        Create the application service.

        @param plugin: The name of the plugin that implements the service
            application to run.
        @param options: Options to pass to the application.
        @return: The created application service.
        """
        service = plugin.makeService(options)
        application = Application(plugin.tapname)
        service.setServiceParent(application)

        return IService(application)

    @staticmethod
    def startService(reactor: IReactorCore, service: IService) -> None:
        """
        Start the application service.

        @param reactor: The reactor to run the service with.
        @param service: The application service to run.
        """
        service.startService()

        # Ask the reactor to stop the service before shutting down
        reactor.addSystemEventTrigger("before", "shutdown", service.stopService)

    @staticmethod
    def run(twistOptions: TwistOptions) -> None:
        """
        Run the application service.

        @param twistOptions: Command line options to convert to runner
            arguments.
        """
        runner = Runner(
            reactor=twistOptions["reactor"],
            defaultLogLevel=twistOptions["logLevel"],
            logFile=twistOptions["logFile"],
            fileLogObserverFactory=twistOptions["fileLogObserverFactory"],
        )
        runner.run()
        reactor = twistOptions["reactor"]
        if _ISupportsExitSignalCapturing.providedBy(reactor):
            if reactor._exitSignal is not None:
                _exitWithSignal(reactor._exitSignal)

    @classmethod
    def main(cls, argv: Sequence[str] = sys.argv) -> None:
        """
        Executable entry point for L{Twist}.
        Processes options and run a twisted reactor with a service.

        @param argv: Command line arguments.
        @type argv: L{list}
        """
        options = cls.options(argv)

        reactor = options["reactor"]
        # If subCommand is None, TwistOptions.parseOptions() raises UsageError
        # and Twist.options() will exit the runner, so we'll never get here.
        subCommand = options.subCommand
        assert subCommand is not None
        service = cls.service(
            plugin=options.plugins[subCommand],
            options=options.subOptions,
        )

        cls.startService(reactor, service)
        cls.run(options)



# -*- test-case-name: twisted.conch.test.test_unix -*-
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.


from zope.interface import implementer

from twisted.conch.interfaces import IConchUser
from twisted.cred.checkers import (
    AllowAnonymousAccess,
    InMemoryUsernamePasswordDatabaseDontUse,
)
from twisted.cred.credentials import (
    Anonymous,
    IAnonymous,
    IUsernamePassword,
    UsernamePassword,
)
from twisted.cred.error import LoginDenied
from twisted.cred.portal import Portal
from twisted.internet.interfaces import IReactorProcess
from twisted.python.fakepwd import UserDatabase
from twisted.python.reflect import requireModule
from twisted.trial import unittest
from .test_session import StubClient, StubConnection

cryptography = requireModule("cryptography")
unix = requireModule("twisted.conch.unix")

if unix is not None:
    from twisted.conch.unix import UnixConchUser, UnixSSHRealm


@implementer(IReactorProcess)
class MockProcessSpawner:
    """
    An L{IReactorProcess} that logs calls to C{spawnProcess}.
    """

    def __init__(self):
        self._spawnProcessCalls = []

    def spawnProcess(
        self,
        processProtocol,
        executable,
        args=(),
        env={},
        path=None,
        uid=None,
        gid=None,
        usePTY=0,
        childFDs=None,
    ):
        """
        Log a call to C{spawnProcess}. Do not actually spawn a process.
        """
        self._spawnProcessCalls.append(
            {
                "processProtocol": processProtocol,
                "executable": executable,
                "args": args,
                "env": env,
                "path": path,
                "uid": uid,
                "gid": gid,
                "usePTY": usePTY,
                "childFDs": childFDs,
            }
        )


shouldSkip = (
    "Cannot run without cryptography"
    if cryptography is None
    else "Unix system required"
    if unix is None
    else None
)


class TestSSHSessionForUnixConchUser(unittest.TestCase):
    skip = shouldSkip

    def testExecCommandEnvironment(self) -> None:
        """
        C{execCommand} sets the C{HOME} environment variable to the avatar's home
        directory.
        """
        userdb = UserDatabase()
        homeDirectory = "/made/up/path/"
        userName = "user"
        userdb.addUser(userName, home=homeDirectory)
        self.patch(unix, "pwd", userdb)
        mockReactor = MockProcessSpawner()
        avatar = UnixConchUser(userName)
        avatar.conn = StubConnection(transport=StubClient())
        session = unix.SSHSessionForUnixConchUser(avatar, reactor=mockReactor)
        protocol = None
        command = ["not-actually-executed"]
        session.execCommand(protocol, command)
        [call] = mockReactor._spawnProcessCalls
        self.assertEqual(homeDirectory, call["env"]["HOME"])


class TestUnixSSHRealm(unittest.TestCase):
    """
    Tests for L{UnixSSHRealm}.
    """

    skip = shouldSkip

    def test_unixSSHRealm(self) -> None:
        """
        L{UnixSSHRealm} is an L{IRealm} whose C{.requestAvatar} method returns
        a L{UnixConchUser}.
        """
        userdb = UserDatabase()
        home = "/testing/home/value"
        userdb.addUser("user", home=home)
        self.patch(unix, "pwd", userdb)
        pwdb = InMemoryUsernamePasswordDatabaseDontUse(user=b"password")
        p = Portal(UnixSSHRealm(), [pwdb])

        # there seems to be a bug in mypy-zope where sometimes things don't
        # implement their superinterfaces; 0.3.11, when we upgrade to 0.9.0
        # this type declaration will be extraneous
        creds: IUsernamePassword = UsernamePassword(b"user", b"password")
        result = p.login(creds, None, IConchUser)
        resultInterface, avatar, logout = self.successResultOf(result)
        self.assertIsInstance(avatar, UnixConchUser)
        assert isinstance(avatar, UnixConchUser)  # legibility for mypy
        self.assertEqual(avatar.getHomeDir(), home)

    def test_unixSSHRefusesAnonymousLogins(self) -> None:
        """
        L{UnixSSHRealm} will refuse anonymous logins.
        """
        p = Portal(UnixSSHRealm(), [AllowAnonymousAccess()])
        result = p.login(IAnonymous(Anonymous()), None, IConchUser)
        loginDenied = self.failureResultOf(result)
        self.assertIsInstance(loginDenied.value, LoginDenied)
